from navix.to_hook_networking import *
from navix.reply_generator import ComputeAndGenerateReply
from navix.monitor import Monitor


#####################################################################
#
# client_process function, handles communication with individual hook
#
#####################################################################


class ClientProcess:
    """
    every connection from an hook to NaviX starts a new process, every process
     initializes its own object from this class.
    """
    def __init__(self, conn, logging, thread_num, settings):
        """
        :param conn: connection to communicate with hook
        :param logging: queue for logging (multiprocessing compatible) -> logging in one file NaviX.log
        :param thread_num: used for identifying process in logs
        :param settings: object containing all important settings
        initializes every used thread global variable and calls afterwards the clientExec
        """
        # list of all attributes
        self.thread_num = thread_num
        self.conn = conn
        self.queue_logging = logging.queue_logging
        self.monitor = Monitor(logging.queue_monitor, logging.queue_logging, thread_num, settings)
        self.data = ""
        self.reply = ""
        self.settings = settings

        # run process
        try:
            self.client_exec()
        except Exception as err:
            try:
                self.queue_logging.put("ERROR: " + str(self.thread_num)
                                       + ": Uncaught exception in client process -> Kill process\n"
                                       + "Python-ERROR: " + str(err))
                sleep(0.1)  # sleep to flush to queue
            except Exception as err:
                print("logger fail!!!!!!\n" + "Python-ERROR: " + str(err))
                exit(-1)
            exit(-1)

        # Function for handling connections. This will be used to create processes
    def client_exec(self):
        """
        :return: nothing -> silent exit
        calls successive functions, so a meaningful answer can be sent back to hook, the steps are:
        # 1 open a private socket for each hook for any further communication
        # 2 receive actual relevant data send by HTCondor via hook
        # 3 get form received data the HPDA_RANK:
        # 3.1 check if HPDA_timestamp in received data -> if yes it is an update hook if not it is a translate hook
        # 3.2 create a list of files specified by user in "input files"
        # 3.4 ask CMS server for the location in file list
        # 3.5 resolve IP given by CMS server to HTCondor name
        # 3.6 create two strings, one for requirement with every server where the job should start and the rank, where
              also the weight of the individual server is considered (if specified)
        # 4 send responds to the hook
        # 5 shut down the individual socket and terminate process
        """
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Thread started")

        #####################################################################
        #
        # open connection on extra port to prevent interference with
        # communication to other hooks
        #
        #####################################################################
        try:
            communication_to_hook = CommunicationToHook(self.conn, self.queue_logging, self.thread_num, self.settings)
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Could not establish connection to hook on 'private' port! Abort!\n"
                                   + "Python-ERROR: " + str(e))
            try:
                communication_to_hook.client_socket.close()
                exit(-1)
            except Exception as e:
                self.queue_logging.put("ERROR: " + str(self.thread_num)
                                       + ": Failed to close socket while aborting! Process further with abort!\n"
                                       + "Python-ERROR: " + str(e))
                sleep(0.1)  # sleep to flush to queue
                exit(-1)

        #####################################################################
        #
        # here starts the communication about the actual information
        # first step: receiving information
        #
        #####################################################################
        try:
            # get data from light_hook
            self.data = communication_to_hook.get_data_send_by_light_hook()
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Failed to receive data from hook!\nAbort process"
                                   + "\nPython-ERROR: " + str(e))
            communication_to_hook.client_conn.close()
            communication_to_hook.client_socket.close()
            self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Client closed")
            sleep(0.1)  # sleep to flush to queue
            exit(-1)

        if self.data == -1:
            try:
                self.queue_logging.put("ERROR: " + str(self.thread_num) + ": No data from lightHook received!\n"
                                       + "Terminating hook and aborting process!")
                self.reply = ("JobStatus = 5\n"
                              + "HoldReason = \"ERROR: No data from lightHook received\"\n"
                              + "HoldReasonCode = 8\n"
                              + "HoldReasonSubCode = 0\n"
                              + "LastJobStatus = 1\n"
                              + "HPDA_REQUIREMENT = false")
                communication_to_hook.send_data_to_light_hook(self.reply)
                communication_to_hook.client_conn.close()
                communication_to_hook.client_socket.close()
                self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Client closed")
                sleep(0.1)  # sleep to flush to queue
                exit(0)
            except Exception as e:
                self.queue_logging.put("ERROR: " + str(self.thread_num)
                                       + ": Failed to send hold signal to condor!\nHard abort process"
                                       + "\nPython-ERROR: " + str(e))
                sleep(0.1)  # sleep to flush to queue
                exit(-1)

        #####################################################################
        #
        # here starts the processing of the received information
        # second step: process information -> check if it is a:
        #               translate_hook: first interaction of hook with NaviX
        #                                -> set requirement and rank
        #               update_hook   : second interaction of hook with NaviX
        #                               -> update requirement and rank if
        #                                  necessary
        #                               -> Job stayed idle for a defined time
        #                                  (time is set in condor)
        #               finalize_hook : last interaction of hook with NaviX
        #                               -> has to return nothing and is only
        #                                  used for monitoring
        #
        #####################################################################
        try:
            # if the connecting hook is the finalize hook, the first line will always be 'FINALIZE'
            # in this case all received data is send to the monitor class and saved to the NaviX.mon file
            # when done with the monitoring log the client process is terminated
            if "FINALIZE" in self.data.splitlines()[0]:
                self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Finalize hook detected!")
                # close connection to client
                communication_to_hook.client_conn.close()
                communication_to_hook.client_socket.close()
                self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Connection closed.")
                self.monitor.finalize(self.data)
                self.queue_logging.put("DEBUG: " + str(self.thread_num)
                                       + ": Done with monitoring / finalizing process!\n")
                sleep(0.1)  # sleep to flush to queue
                exit(0)

            else:
                generate_reply = ComputeAndGenerateReply(self.data, self.queue_logging, self.thread_num, self.settings)
                self.reply = generate_reply.compute_generate_reply()
        except Exception as error:
            try:
                self.queue_logging.put("ERROR: " + str(self.thread_num)
                                       + ": Processing of given data raised an ERROR!\n"
                                       + "Terminating hook and aborting process!"
                                       + "\nPython-ERROR: " + str(error))
                self.reply = ("JobStatus = 5\n"
                              + "HoldReason = \"ERROR: Processing of given data raised an ERROR, "
                              + "please check your InputFiles classAd\"\n"
                              + "HoldReasonCode = 8\n"
                              + "HoldReasonSubCode = 0\n"
                              + "LastJobStatus = 1\n"
                              + "HPDA_REQUIREMENT = false")
                communication_to_hook.send_data_to_light_hook(self.reply)
                communication_to_hook.client_conn.close()
                communication_to_hook.client_socket.close()
                self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Client closed")
                sleep(0.1)  # sleep to flush to queue
                exit(-1)
            except Exception as err:
                self.queue_logging.put("ERROR: " + str(self.thread_num)
                                       + ": Failed to send hold signal to condor!\nHard abort process"
                                       + "\nPython-ERROR: " + str(err))
                sleep(0.1)  # sleep to flush to queue
                exit(-1)
        #####################################################################
        #
        # send reply back to the translate/update hook
        # -> reply is a simple string send back over tcp
        # -> terminate client process afterwards
        #
        #####################################################################
        try:
            communication_to_hook.send_data_to_light_hook(self.reply)
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num) + ": Failed sending generated reply to hook!"
                                   + "\nPython-ERROR: " + str(e))
        # close connection to client
        try:
            communication_to_hook.client_conn.close()
            communication_to_hook.client_socket.close()
            self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Connection closed")
        except Exception as err:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Failed to close connection to hook!\nExit process!"
                                   + "\nPython-ERROR: " + str(err))
            sleep(0.1)  # sleep to flush to queue
            exit(-1)
        sleep(0.1)  # sleep to flush to queue
        exit(0)
