import logging.config
import logging
from logging.handlers import RotatingFileHandler
from multiprocessing import Process, Queue
from navix.settings import Settings


class Logging:
    """"
    class contains logging function for normal logging and monitoring logging
    logging is set up with a multiprocessing queue to be compatible with multiprocessing
    """
    def __init__(self):
        #####################################################################
        #
        # Start logger => errors above will be logged with start_log
        # Afterwards the start_log should not be used any further, otherwise
        # file corruption is possible
        #
        #####################################################################

        # logging queue
        self.queue_logging = Queue()
        self.queue_logging.cancel_join_thread()
        self.queue_monitor = Queue()
        self.queue_monitor.cancel_join_thread()

        # initialize and start logging process
        log_writer = Process(target=logger, args=(self.queue_logging,))
        log_writer.daemon = True
        log_writer.start()

        # initialize and start monitor_logging process
        monitor_log_writer = Process(target=monitor, args=(self.queue_monitor,))
        monitor_log_writer.daemon = True
        monitor_log_writer.start()


#####################################################################
#
# Configure and start queue for logging
# (queue is used because logging is only thread safe
#  but not multiprocessing
#  -> write all to one queue, read by one thread an put it into the
#     logfile
# )
#
#####################################################################


def configure_logger():
    # log file path
    log_file = 'NaviX.log'

    # create logger
    log = logging.getLogger(log_file)

    settings = Settings.get_instance(log)

    log.setLevel(logging.getLevelName(settings.log_level))


    # create console handler and set level to debug
    console_log = logging.StreamHandler()

    # create formatter
    formatter_log = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y/%m/%d %H:%M:%S')

    # add formatter to ch
    console_log.setFormatter(formatter_log)

    # add ch to logger
    log.addHandler(console_log)
    log_handler = RotatingFileHandler(log_file, mode='a', maxBytes=1 * 1024 * 1024 * 1024,
                                      backupCount=15)
    log_handler.setFormatter(formatter_log)
    log.addHandler(log_handler)
    return log


def configure_monitor_logger():
    # monitor file path
    monitor_logger_file = 'NaviX.mon'

    # create monitoring logger
    monitor_logger = logging.getLogger(monitor_logger_file)
    monitor_logger.setLevel(logging.INFO)

    # file rotation
    monitor_handler = RotatingFileHandler(monitor_logger_file, mode='a', maxBytes=50 * 1024 * 1024,
                                          backupCount=90)

    # create formatter
    formatter_monitor_logger = logging.Formatter('%(message)s')
    monitor_handler.setFormatter(formatter_monitor_logger)

    # set to monitor logger
    monitor_logger.addHandler(monitor_handler)
    return monitor_logger


def logger(queue):
    # initialize logger
    log = configure_logger()
    while True:
        msg = queue.get()
        if "DEBUG: " == msg[:7]:
            log.debug(msg[7:])
        elif "_INFO: " == msg[:7]:
            log.info(msg[7:])
        elif "ERROR: " == msg[:7]:
            log.error(msg[7:])
        elif "WARNI: " == msg[:7]:
            log.warning(msg[7:])
        else:
            log.error(msg)


def monitor(queue):
    # initialize monitor_logger
    log = configure_monitor_logger()
    while True:
        msg = queue.get()
        log.info(msg)
