import configparser
import yaml
import ipaddress
##########################################################
#
# Variable Block with default values
# -> main function reads from config file and updates them
#
##########################################################


class Settings(object):
    __instance = None

    ##########################################################
    #
    # config is programmed to be a singleton object
    #
    ##########################################################
    @staticmethod
    def get_instance(logging, logging_level=False):
        """
        :param logging:
        :param logging_level:
        :return:
        Create instance of Settings if it does not exist. Afterwards, return the instance.
        """
        # Static access method
        if Settings.__instance is None:
            if logging_level not in ["DEBUG", "INFO", False]:
                logging_level = False
                logging.error("Logging level given as argument could not be identified and is therefor ignored!\n"
                              + "Given logging level was " + str(logging_level))
            Settings(logging, logging_level)
        return Settings.__instance

    def __init__(self, logging, logging_level = False):
        """
        :param logging:
        :param logging_level:
        :return:
        Initialize instance of Settings with default values.
        """
        # this will raise an error if someone tries to initialize a new object, if there is already one
        if Settings.__instance is not None:
            raise Exception("Settings was read a second time! This should have not happened!")
        else:
            ##########################################################
            #
            # stores all global config
            #
            ##########################################################
            self.__HOST = ''  # Symbolic name meaning all available interfaces
            self.__PORT = 8889  # Arbitrary non-privileged port
            self.__termination_signal = "END11111111111111111111111111111111111111111111111111111111111" \
                                        "1111111111111111111111111111111111111111"
            # xrootd_server_address = "root://localhost:1094"#"root://localhost:1234"
            self.__xrootd_server_address = "root://condorcentral.etp.kit.edu:1094"
            self.__server_weight = 100
            self.__pyxrootd_timeout = 20
            self.__known_servers = dict()
            self.__compile_hook = True
            self.__source_path_translate_hook = "hooks/lightHook_translate.cpp"
            self.__target_compile_translate_hook = "hooks/lightHook_translate"
            self.__source_path_update_hook = "hooks/lightHook_update.cpp"
            self.__target_compile_update_hook = "hooks/lightHook_update"
            self.__source_path_finalize_hook = "hooks/lightHook_finalize.cpp"
            self.__target_compile_finalize_hook = "hooks/lightHook_finalize"
            self.__source_path_hook_main = "hooks/main.h"
            self.__time_before_adding_additional_resources = 360000
            self.__cached_percentage_before_adding_to_requirement = 0.5
            self.__log_level = "Debug"
            self.__working_directory = "Iwd"
            self.__requested_files = "Input_Files"
            self.__requested_files_tmp = "InputFiles_tmp"
            self.__Requested_files_mon = "Requested_files"
            self.__HPDA_RANK = "HPDA_RANK"
            self.__HPDA_REQUIREMENT = "HPDA_REQUIREMENT"
            self.__Rank = "RANK"
            self.__HPDA_TIMESTAMP = "HPDA_TIMESTAMP"
            self.__start_job_timestamp = "JobCurrentStartDate"
            self.__completion_timestamp = "CompletionDate"
            self.__File_found_on_server = "File_found_on_server"
            self.__Requested_files = "Requested_files"
            self.__HPDA_RANK_MONITORING = "HPDA_RANK_MONITORING"
            self.__GlobalJobId = "GlobalJobId"
            self.__RemoteUserCpu = "RemoteUserCpu"
            self.__RemoteSysCpu = "RemoteSysCpu"
            self.__RequestCpus = "RequestCpus"
            self.__NetworkInputMb = "NetworkInputMb"
            self.__LastRemoteHost = "LastRemoteHost"
            self.__LastRemoteIP = "StartdPrincipal"

            if self.read_navix_config(logging, logging_level) == -1:
                raise Exception("Failed to get config!")
            if self.read_htcondor_bindings_config(logging) == -1:
                raise Exception("Failed to get HTCondor bindings!")
            if self.read_server_list(logging) == -1:
                raise Exception("Failed to read known server!")

            Settings.__instance = self

    #######################
    #
    # getter block
    #
    #######################
    def __get_host(self):
        return self.__HOST

    def __get_port(self):
        return self.__PORT

    def __get_termination_signal(self):
        return self.__termination_signal

    def __get_xrootd_server_address(self):
        return self.__xrootd_server_address

    def __get_server_weight(self):
        return self.__server_weight

    def __get_pyxrootd_timeout(self):
        return self.__pyxrootd_timeout

    def __get_known_servers(self):
        return self.__known_servers

    def __get_compile_hook(self):
        return self.__compile_hook

    def __get_source_path_translate_hook(self):
        return self.__source_path_translate_hook

    def __get_target_compile_translate_hook(self):
        return self.__target_compile_translate_hook

    def __get_source_path_update_hook(self):
        return self.__source_path_update_hook

    def __get_target_compile_update_hook(self):
        return self.__target_compile_update_hook

    def __get_source_path_finalize_hook(self):
        return self.__source_path_finalize_hook

    def __get_target_compile_finalize_hook(self):
        return self.__target_compile_finalize_hook

    def __get_source_path_hook_main(self):
        return self.__source_path_hook_main

    def __get_time_before_adding_additional_resources(self):
        return self.__time_before_adding_additional_resources

    def __get_log_level(self):
        return self.__log_level

    def __get_cached_percentage_before_adding_to_requirement(self):
        return self.__cached_percentage_before_adding_to_requirement

    def __get_working_directory(self):
        return self.__working_directory

    def __get_requested_files(self):
        return self.__requested_files

    def __get_requested_files_tmp(self):
        return self.__requested_files_tmp

    def __get_requested_files_mon(self):
        return self.__Requested_files_mon

    def __get_hpda_rank(self):
        return self.__HPDA_RANK

    def __get_hpda_requirement(self):
        return self.__HPDA_REQUIREMENT

    def __get_rank(self):
        return self.__Rank

    def __get_hpda_timestamp(self):
        return self.__HPDA_TIMESTAMP

    def __get_start_job_timestamp(self):
        return self.__start_job_timestamp

    def __get_completion_timestamp(self):
        return self.__completion_timestamp

    def __get_file_found_on_server(self):
        return self.__File_found_on_server

    def __get_hpda_rank_monitoring(self):
        return self.__HPDA_RANK_MONITORING

    def __get_global_job_id(self):
        return self.__GlobalJobId

    def __get_remote_user_cpu(self):
        return self.__RemoteUserCpu

    def __get_remote_sys_cpu(self):
        return self.__RemoteSysCpu

    def __get_request_cpus(self):
        return self.__RequestCpus

    def __get_network_input_mb(self):
        return self.__NetworkInputMb

    def __get_last_remote_host(self):
        return self.__LastRemoteHost

    def __get_last_remote_ip(self):
        return self.__LastRemoteIP

    host = property(__get_host)
    port = property(__get_port)
    termination_signal = property(__get_termination_signal)
    xrootd_server_address = property(__get_xrootd_server_address)
    server_weight = property(__get_server_weight)
    pyxrootd_timeout = property(__get_pyxrootd_timeout)
    known_servers = property(__get_known_servers)
    compile_hook = property(__get_compile_hook)
    source_path_translate_hook = property(__get_source_path_translate_hook)
    target_compile_translate_hook = property(__get_target_compile_translate_hook)
    source_path_update_hook = property(__get_source_path_update_hook)
    target_compile_update_hook = property(__get_target_compile_update_hook)
    source_path_finalize_hook = property(__get_source_path_finalize_hook)
    target_compile_finalize_hook = property(__get_target_compile_finalize_hook)
    source_path_hook_main = property(__get_source_path_hook_main)
    time_before_adding_additional_resources = property(__get_time_before_adding_additional_resources)
    log_level = property(__get_log_level)
    cached_percentage_before_adding_to_requirement = property(__get_cached_percentage_before_adding_to_requirement)
    working_directory = property(__get_working_directory)
    requested_files = property(__get_requested_files)
    requested_files_tmp = property(__get_requested_files_tmp)
    requested_files_mon = property(__get_requested_files_mon)
    hpda_rank = property(__get_hpda_rank)
    hpda_requirement = property(__get_hpda_requirement)
    rank = property(__get_rank)
    hpda_timestamp = property(__get_hpda_timestamp)
    start_job_timestamp = property(__get_start_job_timestamp)
    completion_timestamp = property(__get_completion_timestamp)
    file_found_on_server = property(__get_file_found_on_server)
    hpda_rank_monitoring = property(__get_hpda_rank_monitoring)
    global_job_id = property(__get_global_job_id)
    remote_user_cpu = property(__get_remote_user_cpu)
    remote_sys_cpu = property(__get_remote_sys_cpu)
    requested_cpus = property(__get_request_cpus)
    network_input_mb = property(__get_network_input_mb)
    last_remote_host = property(__get_last_remote_host)
    last_remote_ip = property(__get_last_remote_ip)

    #######################
    #
    # read configs
    #
    #######################

    def read_navix_config(self, logging, logging_level):
        """
        :param logging:
        :param logging_level:
        :return:
        sets the config given in the file "NaviX.cnf".
        if a parameter is not set or not found, e.g. due typo, the value falls back to the default
        given in NaviX_settings and logs a warning all config config are saved in NaviX_settings
        """
        try:
            config_path = "config/NaviX.cnf"
            logging.debug("Reading config from: " + config_path)
            config = configparser.ConfigParser()
            exists_check = config.read(config_path)
            if not exists_check:
                raise Exception("Config file not found!")
            try:
                self.__pyxrootd_timeout = int(config.get('main', 'pyxrootd_timeout'))
                logging.debug("Pyxrootd timeout set to: " + str(self.__pyxrootd_timeout))
            except Exception as err:
                logging.info("Failed to get timeout for pyxrootd from config, using: "
                             + str(self.pyxrootd_timeout) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__server_weight = int(config.get('main', 'server_weight'))
                logging.debug("Server default weight set to: " + str(self.__server_weight))
            except Exception as err:
                logging.info("Failed to get server_weight from config, using: "
                             + str(self.server_weight) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__xrootd_server_address = config.get('main', 'xrootd_server_address')
                logging.debug("Xrootd server address is set to: " + str(self.__xrootd_server_address))
            except Exception as err:
                logging.info("Failed to get xrootd server address from config, using: "
                             + str(self.__xrootd_server_address) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__PORT = int(config.get('main', 'PORT'))
                logging.debug("NaviX port is set to: " + str(self.__PORT))
            except Exception as err:
                logging.info("Failed to get port for NaviX from config, using: "
                             + str(self.__PORT) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__compile_hook = bool(eval(config.get('main', 'compile_hook')))
                logging.debug("Compile hook on startup: " + str(self.__compile_hook))
            except Exception as err:
                logging.info("Failed to get if code should compile the hook on startup from config, using: "
                             + str(self.__compile_hook) + " as default value!\n"
                             + "Python-Error: " + str(err))
            try:
                self.__source_path_translate_hook = config.get('main', 'source_path_translate_hook')
                logging.debug("Source to compile translate hook is: " + str(self.__source_path_translate_hook))
            except Exception as err:
                logging.info(
                    "Failed to get the path to the source of the translate hook on startup from config, using: "
                    + str(self.__source_path_translate_hook) + " as default value!\n"
                    + "Python-Error: " + str(err))
            try:
                self.__target_compile_translate_hook = config.get('main', 'target_compile_translate_hook')
                logging.debug("Place where to place the executable translate hook: "
                              + str(self.__target_compile_translate_hook))
            except Exception as err:
                logging.info("Failed to get the path where to output compiled translate hook on startup from config, "
                             + "using: " + str(self.__target_compile_translate_hook) + " as default value!\n"
                             + "Python-Error: " + str(err))
            try:
                self.__source_path_update_hook = config.get('main', 'source_path_update_hook')
                logging.debug("Source to compile update hook is: " + str(self.__source_path_update_hook))
            except Exception as err:
                logging.info("Failed to get the path to the source of the update hook on startup from config, using: "
                             + str(self.__source_path_update_hook) + " as default value!\n"
                             + "Python-Error: " + str(err))
            try:
                self.__target_compile_update_hook = config.get('main', 'target_compile_update_hook')
                logging.debug("Place where to place the executable update hook: "
                              + str(self.__target_compile_update_hook))
            except Exception as err:
                logging.info(
                    "Failed to get the path where to output compiled update hook on startup from config, using: "
                    + str(self.__target_compile_update_hook) + " as default value!\n"
                    + "Python-Error: " + str(err))
            try:
                self.__source_path_finalize_hook = config.get('main', 'source_path_finalize_hook')
                logging.debug("Source to compile finalize hook is: " + str(self.__source_path_finalize_hook))
            except Exception as err:
                logging.info("Failed to get the path to the source of the finalize hook on startup from config, using: "
                             + str(self.__source_path_finalize_hook) + " as default value!\n"
                             + "Python-Error: " + str(err))
            try:
                self.__target_compile_finalize_hook = config.get('main', 'target_compile_finalize_hook')
                logging.debug("Place where to place the executable finalize hook: "
                              + str(self.__target_compile_update_hook))
            except Exception as err:
                logging.info(
                    "Failed to get the path where to output compiled finalize hook on startup from config, using: "
                    + str(self.__target_compile_finalize_hook) + " as default value!\n"
                    + "Python-Error: " + str(err))
            try:
                self.__source_path_hook_main = config.get('main', 'source_path_hook_main')
                logging.debug("Source of main file for all hooks to compile is: "
                              + str(self.__source_path_hook_main))
            except Exception as err:
                logging.info(
                    "Failed to get the to the source of the main file for all hooks on startup from config, using: "
                    + str(self.__source_path_hook_main) + " as default value!\n"
                    + "Python-Error: " + str(err))
            try:
                if logging_level:
                    self.__log_level = logging_level
                    logging.debug("Log level given as argument is: " + str(self.__log_level))
                else:
                    self.__log_level = config.get('main', 'log_level')
                    logging.debug("Log level read from config is: " + str(self.__log_level))
            except Exception as err:
                logging.info("Failed to get log level, using: " + str(self.__log_level) + " as default value!\n"
                             + "Python-Error: " + str(err))
            try:
                self.__time_before_adding_additional_resources = \
                    int(config.get('main', 'time_before_adding_additional_resources'))
                logging.debug("Time before new resources are added is set to: "
                              + str(self.__time_before_adding_additional_resources))
            except Exception as err:
                logging.info("Failed to get time before new resources are added, using: "
                             + str(self.__time_before_adding_additional_resources) + " as default value!\n"
                             + "Python-Error: " + str(err))
            try:
                self.__cached_percentage_before_adding_to_requirement = \
                    float(config.get('main', 'cached_percentage_before_adding_to_requirement'))
                logging.debug("Percentage of cached files before a node is added is set to: "
                              + str(self.__cached_percentage_before_adding_to_requirement))
            except Exception as err:
                logging.info("Failed to get percentage of cached files before a node is added, using: "
                             + str(self.__cached_percentage_before_adding_to_requirement) + " as default value!\n"
                             + "Python-Error: " + str(err))

        except Exception as err:
            logging.error("Could not read config file!\n\nPlease edit 'NaviX.cnf' to prevent this Error"
                          "Python-Error: " + str(err))
            raise

    def read_htcondor_bindings_config(self, logging):
        """
         :param logging:
         :return:
         sets the HTCondor bindings given in the file "HTCondor_bindings.cnf".
         if a parameter is not set or not found, e.g. due typo, the value falls back to the default
         given in NaviX_settings and logs a warning all config config are saved in NaviX_settings
         """
        try:
            config_path = "config/HTCondor_bindings.cnf"
            logging.debug("Reading config from: " + config_path)
            config = configparser.ConfigParser()
            exists_check = config.read(config_path)
            if not exists_check:
                raise Exception("HTCondor_bindings file not found!")

            try:
                self.__working_directory = str(config.get('condor_bindings', 'working_directory'))
                logging.debug("HTCondor binding: working directory set to: " + str(self.__working_directory))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for working directory from config, using: "
                             + str(self.working_directory) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__requested_files = str(config.get('condor_bindings', 'requested_files'))
                logging.debug("HTCondor binding: requested_files set to: " + str(self.requested_files))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for requested_files from config, using: "
                             + str(self.requested_files) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__Requested_files_mon = str(config.get('condor_bindings', 'Requested_files_mon'))
                logging.debug("HTCondor binding: Requested_files_mon set to: " + str(self.requested_files_mon))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for Requested_files_mon from config, using: "
                             + str(self.requested_files_mon) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__HPDA_RANK = str(config.get('condor_bindings', 'HPDA_RANK'))
                logging.debug("HTCondor binding: HPDA_RANK set to: " + str(self.hpda_rank))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for HPDA_RANK from config, using: "
                             + str(self.hpda_rank) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__HPDA_REQUIREMENT = str(config.get('condor_bindings', 'HPDA_REQUIREMENT'))
                logging.debug("HTCondor binding: HPDA_REQUIREMENT set to: " + str(self.hpda_requirement))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for HPDA_REQUIREMENT from config, using: "
                             + str(self.hpda_requirement) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__Rank = str(config.get('condor_bindings', 'Rank'))
                logging.debug("HTCondor binding: RANK set to: " + str(self.rank))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for RANK from config, using: "
                             + str(self.rank) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__HPDA_TIMESTAMP = str(config.get('condor_bindings', 'HPDA_TIMESTAMP'))
                logging.debug("HTCondor binding: HPDA_TIMESTAMP set to: " + str(self.hpda_timestamp))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for HPDA_TIMESTAMP from config, using: "
                             + str(self.hpda_timestamp) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__start_job_timestamp = str(config.get('condor_bindings', 'start_job_timestamp'))
                logging.debug("HTCondor binding: start_job_timestamp set to: " + str(self.start_job_timestamp))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for start_job_timestamp from config, using: "
                             + str(self.start_job_timestamp) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__completion_timestamp = str(config.get('condor_bindings', 'completion_timestamp'))
                logging.debug("HTCondor binding: completion_timestamp set to: " + str(self.completion_timestamp))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for completion_timestamp from config, using: "
                             + str(self.completion_timestamp) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__File_found_on_server = str(config.get('condor_bindings', 'File_found_on_server'))
                logging.debug("HTCondor binding: File_found_on_server set to: " + str(self.file_found_on_server))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for File_found_on_server from config, using: "
                             + str(self.file_found_on_server) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__Requested_files = str(config.get('condor_bindings', 'Requested_files'))
                logging.debug("HTCondor binding: Requested_files set to: " + str(self.requested_files))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for Requested_files from config, using: "
                             + str(self.requested_files) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__HPDA_RANK_MONITORING = str(config.get('condor_bindings', 'HPDA_RANK_MONITORING'))
                logging.debug("HTCondor binding: HPDA_RANK_MONITORING set to: " + str(self.hpda_rank_monitoring))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for HPDA_RANK_MONITORING from config, using: "
                             + str(self.hpda_rank_monitoring) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__GlobalJobId = str(config.get('condor_bindings', 'GlobalJobId'))
                logging.debug("HTCondor binding: GlobalJobId set to: " + str(self.global_job_id))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for GlobalJobId from config, using: "
                             + str(self.global_job_id) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__RemoteUserCpu = str(config.get('condor_bindings', 'RemoteUserCpu'))
                logging.debug("HTCondor binding: RemoteUserCpu set to: " + str(self.remote_user_cpu))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for RemoteUserCpu from config, using: "
                             + str(self.remote_user_cpu) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__RemoteSysCpu = str(config.get('condor_bindings', 'RemoteSysCpu'))
                logging.debug("HTCondor binding: RemoteSysCpu set to: " + str(self.remote_sys_cpu))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for RemoteSysCpu from config, using: "
                             + str(self.remote_sys_cpu) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__RequestCpus = str(config.get('condor_bindings', 'RequestCpus'))
                logging.debug("HTCondor binding: RequestCpus set to: " + str(self.requested_cpus))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for RequestCpus from config, using: "
                             + str(self.requested_cpus) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__NetworkInputMb = str(config.get('condor_bindings', 'NetworkInputMb'))
                logging.debug("HTCondor binding: NetworkInputMb set to: " + str(self.network_input_mb))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for NetworkInputMb from config, using: "
                             + str(self.network_input_mb) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__LastRemoteHost = str(config.get('condor_bindings', 'LastRemoteHost'))
                logging.debug("HTCondor binding: LastRemoteHost set to: " + str(self.last_remote_host))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for LastRemoteHost from config, using: "
                             + str(self.last_remote_host) + " as default value!"
                             + "Python-Error: " + str(err))
            try:
                self.__LastRemoteIP = str(config.get('condor_bindings', 'LastRemoteIP'))
                logging.debug("HTCondor binding: LastRemoteIP set to: " + str(self.last_remote_ip))
            except Exception as err:
                logging.info("Failed to get HTCondor binding for LastRemoteIP from config, using: "
                             + str(self.last_remote_ip) + " as default value!"
                             + "Python-Error: " + str(err))
        except Exception as err:
            logging.error("Could not read config file!\n\nPlease edit 'HTCondor_bindings.cnf' to prevent this Error"
                          "Python-Error: " + str(err))
            raise

    def read_server_list(self, logging):
        """
        :param logging::
        :return:
        the server list is read from "known_servers.yml", which is a yaml file. The IPs are checked if it is a valid
        entry and if it is an IPv4 address it is transformed to IPv6
        Neither the server weight nor the the name are checked! The name should be known by HTCondor!
        The server prints the given server to the log, so every user can check if the servers are read well.
        """
        logging.debug("Start reading server list.")
        ##########################################################
        # Read config file
        # to resolve IP Addresses to HTCondor server names
        ##########################################################
        try:
            with open('config/known_servers.yml') as config_file:
                self.__known_servers = yaml.safe_load(config_file)
            config_file.close()
            logging.debug("'config/known_servers.yml' found and opened")
        except Exception as e:
            logging.error("Could not read config \"known_servers.yml\" -> stopping server\nPython-ERROR: " + str(e))
            raise
        try:
            # Deepcopy for iterating
            servers = self.__known_servers.copy()
        except Exception as e:
            logging.error("Config seems to be empty -> stopping server\nPython-ERROR: " + str(e)
                          + "\nPlease edit 'known_servers.yml' to prevent this Error")
            raise
        try:
            # Test given IPs
            for IP in servers.keys():
                if self.test_and_transform_ip(IP, logging) == -1:
                    del self.__known_servers[IP]
                    logging.warning("Given IP " + str(IP) + " was removed.")

            '#Show given hosts to terminal and logfile'
            logging.debug("Known hosts:")
            for IP, server in iter(self.__known_servers.items()):
                if server.get('weight') is not None:
                    logging.debug(
                        str(server.get('name')) + ": " + str(IP) + " => weight: " + str(server.get('weight')))
                else:
                    logging.debug(str(server.get('name')) + ": " + str(IP) + " => weight: " + "default")
        except Exception as e:
            logging.error("No servers found -> stopping server\nPython-ERROR: " + str(e)
                          + "\nPlease edit 'known_servers.yml' to prevent this Error")
            raise

    def test_and_transform_ip(self, ip_in, logging):
        """
        :param ip_in:
        :param logging:
        :return:
        here every given IP or IP-range is tested, if it is a valid IPv6 or IPv4 address, if it is a valid IPv4 address
        it will be transformed to an IPv6 address. The old IPv4 entry in NaviX_settings will be replaced by an
        IPv6 entry.
        """
        ip = ip_in
        try:
            if str(ipaddress.ip_network(ip_in)) != str(ip_in):
                logging.debug("Given IP " + str(ip_in) + " will be set to " + str(ipaddress.ip_network(ip_in)))
                ip = str(ipaddress.ip_network(ip_in))
            logging.debug("Given IP " + str(ip_in) + " is valid.")
        except Exception as e:
            logging.warning("Given IP " + str(ip_in) + " is not valid, \nPython-ERROR: " + str(e))
            return -1
        try:  # IP4 to IPV6
            if ipaddress.ip_address((ip.split("/")[0])).version == 4:
                ip = "::" + str(ip.split("/")[0]) + "/" + str(96 + int(ip.split("/")[1]))
                logging.debug("Given IP was successfully transformed to IPv6: " + str(ip))
            self.__known_servers[ip] = self.__known_servers.pop(ip_in)
            return 0
        except Exception as e:
            logging.warning("Given IP " + str(ip) + " could not be transformed to IPv6 \nPython-ERROR: " + str(e))
            return -1

    def _drop(self):
        # Drop the instance (for testing purposes only!)
        Settings.__instance = None
