
from time import time, sleep
import socket
import sys

###################################################################################################################
#
# This function and class is needed for communication between hooks to NaviX
#
###################################################################################################################


def open_socket(port, queue_logging, process, host):
    """
    :param port:
    :param queue_logging:
    :param process:
    :param host:
    :return:
    function to open socket for the server itself and every thread
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    queue_logging.put("DEBUG: " + str(process) + ": Socket created")
    # Bind socket to local host and port
    try:
        sock.bind((host, port))
    except socket.error as msg:
        queue_logging.put("ERROR: " + 'Bind failed. Error Code : ' + str(msg[0]) + ' Message ' + msg[1])
        sys.exit()
    queue_logging.put("DEBUG: " + str(process) + ": Socket bind complete")
    # Start listening on socket
    sock.listen(2500)
    queue_logging.put("DEBUG: " + str(process) + ": Socket now listening")
    return sock


class CommunicationToHook:

    def __init__(self, conn, queue_logging, thread_num, settings):
        self.thread_num = thread_num
        self.conn = conn
        self.queue_logging = queue_logging
        self.start_time = time()
        self.client_socket = 1
        self.client_conn = -1
        self.settings = settings
        if self.start_connection() == -1:
            raise ValueError('Failed to open private socket for hook!')
        if self.establish_private_port_communication_to_hook() == -1:
            raise ValueError('Hook failed with establishing a connection to private port !')

    def start_connection(self):
        try:
            self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Opening client socket")
            # open new client socket
            self.client_socket = open_socket(0, self.queue_logging, self.thread_num, self.settings.host)
        except Exception as e:
            self.queue_logging.put("DEBUG: " + str(self.thread_num)
                                   + ": Failed opening client socket -> Abort thread\n\n"
                                   + "Maybe there is no free port on host machine available?"
                                   + "\nPython-ERROR: " + str(e))
            raise ValueError("Failed to start connection!")

    def get_data_send_by_light_hook(self):
        """
        :param self:
        :return:
        Function to receive data form light hook via socket
        """
        input_string_from_hook = ""
        # infinite loop so that function do not terminate and thread do not end
        # -> except if termination signal arrives -> break loop
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Start receiving actual data from hook")
        while True:
            # Receiving from client
            data_received = self.client_conn.recv(1024)
            input_string_from_hook = input_string_from_hook + data_received.decode('utf-8')
            if self.settings.termination_signal in str(input_string_from_hook):
                input_string_from_hook = input_string_from_hook.replace(self.settings.termination_signal, "")
                self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Data received from hook: "
                                       + input_string_from_hook)
                return input_string_from_hook
            # timeout on broken pipe ~20secs
            if time() - self.start_time > 20:
                self.queue_logging.put("WARNI: " + str(self.thread_num) + ": Communication with hook ran into timeout!")
                return -1

    def send_data_to_light_hook(self, reply):
        """
        :param self:
        :param reply:
        :return:
        Function to send data to light hook via socket
        """
        # Display reply on server output
        self.queue_logging.put("DEBUG: " + str(self.thread_num)
                               + ": Reply:\n"
                               + "\n###########################################################\n"
                               + str(reply)
                               + " \nSending reply\n"
                               + "###########################################################")
        # came out of loop => return via socket
        # Sending line per line -> if all data is send at once some wired chars appear from time to time or only one
        # half is transmitted
        for line in reply.splitlines():
            send_package = [line[i:i + 1022] for i in range(0, len(line), 1022)]
            for package in send_package:
                self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Sending bytes: "
                                       + str(len(package.encode('utf-8')) + 1) + " Text is: " + package)
                # send packages of line
                if package is send_package[-1]:
                    self.client_conn.send(bytes(("1" + package + "\n"), 'utf-8'))
                else:
                    self.client_conn.send(bytes(("0" + package), 'utf-8'))
                # sleep is needed to separate lines
                sleep(0.1)
        self.client_conn.send(bytes(self.settings.termination_signal, 'utf-8'))
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Done... next step shutdown of thread")

    def establish_private_port_communication_to_hook(self):
        """
        :param self:
        :return:
        function to establish communication to a hook on a separate port
        (as identifier the pid of the hook is send back)
        """
        try:
            data_received = self.conn.recv(1024)
            # Sending message to connected client, -> giving new individual port to connect
            self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Data received by hook:"
                                   + str(data_received.decode("utf-8"))
                                   + ":" + str(self.client_socket.getsockname()[1]))
            self.conn.send(bytes((data_received.decode("utf-8") + ":" + str(self.client_socket.getsockname()[1])),
                                 'utf-8'))
            # send only takes bitstring
        except Exception as err:
            self.queue_logging.put("DEBUG: " + str(self.thread_num)
                                   + ": Failed to communicate with hook! -> Abort thread! \n ERROR given: "
                                   + str(err))
            self.client_socket.close()
            self.conn.close()
            return -1

        # wait for client to connect to new port
        try:
            self.client_socket.settimeout(60)
            self.client_conn, client_address = self.client_socket.accept()
        except Exception as e:
            self.queue_logging.put("ERROR: "
                                   + str(self.thread_num)
                                   + ":client did not successfully connect to individual port on host"
                                   + "\nPython-ERROR: " + str(e))
            self.client_socket.close()
            self.conn.close()
            return -1
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Connected with client on individual port: "
                               + str(client_address[0]) + ':' + str(client_address[1]))
