from unittest import TestCase
from navix.logging import *
from navix.settings import Settings
from os import rename, mknod
from time import sleep


class TestSettings(TestCase):
    __instance = None
    maxDiff = None

    def setUp(self):
        # log file path
        self.log_file_s = 'NaviX_start_error.log2'
        # create logger
        start_log = logging.getLogger(self.log_file_s)
        if TestSettings.__instance is None:
            TestSettings.__instance = 1
            start_log.setLevel(logging.DEBUG)
            handler = logging.handlers.RotatingFileHandler(self.log_file_s, mode='a')
            start_log.addHandler(handler)

        self.thread_num = "TEST"
        self.logger = start_log
        return

    def test_get_instance(self):
        Settings.get_instance(self.logger)
        with self.assertRaises(Exception) as err:
            Settings(self.logger)
            self.assertEqual("Settings was read a second time! This should have not happened!", str(err))
        settings = Settings.get_instance(self.logger)
        self.assertEqual("", settings.host)
        self.assertEqual(8889, settings.port)

    def test_singleton(self):
        Settings.get_instance(self.logger)
        with self.assertRaises(Exception) as err:
            Settings(self.logger)
        self.assertEqual("Settings was read a second time! This should have not happened!", str(err.exception))

    def test_missing_navix_config(self):
        # Check missing NaviX.cnf
        Settings.get_instance(self.logger)._drop()
        rename("config/NaviX.cnf", "config/NaviX.cnf.bak")
        try:
            with self.assertRaises(Exception):
                settings = Settings.get_instance(self.logger)
                settings.read_navix_config(self.logger, "DEBUG")
            sleep(1) # Time for logging
            with open(self.log_file_s, "r") as log_file:
                lines = log_file.readlines()  # Read to end
            log_file.close()
            with open('tests/test_configs/expected_log_no_navix_cnf_file.txt', "r") as log_file:
                ref_lines = log_file.readlines()  # Read to end
            log_file.close()
            lines = lines[-(len(ref_lines)):]
            # Test log output
            self.assertListEqual(sorted(ref_lines), sorted(lines))
            # Test return value
        finally:
            rename("config/NaviX.cnf.bak", "config/NaviX.cnf")

    def test_missing_known_servers_config(self):
        # Check missing known_servers.yml
        Settings.get_instance(self.logger)._drop()
        rename("config/known_servers.yml", "config/known_servers.yml.bak")
        try:
            with self.assertRaises(Exception):
                settings = Settings.get_instance(self.logger)
                settings.read_server_list(self.logger)
            sleep(1)  # Time for logging
            with open(self.log_file_s, "r") as log_file:
                lines = log_file.readlines()  # Read to end
            log_file.close()
            with open('tests/test_configs/expected_log_no_known_servers_cnf_file.txt', "r") as log_file:
                ref_lines = log_file.readlines()  # Read to end
            log_file.close()
            lines = lines[-(len(ref_lines)):]
            # Test log output
            self.assertListEqual(sorted(ref_lines), sorted(lines))
        finally:
            rename("config/known_servers.yml.bak", "config/known_servers.yml")

    def test_missing_htcondor_bindings_config(self):
        # Check missing HTCondor_bindings.cnf
        Settings.get_instance(self.logger)._drop()
        rename("config/HTCondor_bindings.cnf", "config/HTCondor_bindings.cnf.bak")
        try:
            with self.assertRaises(Exception):
                settings = Settings.get_instance(self.logger)
                settings.read_htcondor_bindings_config(self.logger)
            sleep(1)  # Time for logging
            with open(self.log_file_s, "r") as log_file:
                lines = log_file.readlines()  # Read to end
            log_file.close()
            with open('tests/test_configs/expected_log_no_htcondor_bindings_cnf_file.txt', "r") as log_file:
                ref_lines = log_file.readlines()  # Read to end
            log_file.close()
            lines = lines[-(len(ref_lines)):]
            # Test log output
            self.assertListEqual(sorted(ref_lines), sorted(lines))
        finally:
            rename("config/HTCondor_bindings.cnf.bak", "config/HTCondor_bindings.cnf")

    def test_empty_navix_config(self):
        # Check empty NaviX.cnf
        Settings.get_instance(self.logger)._drop()
        rename("config/NaviX.cnf", "config/NaviX.cnf.bak")
        try:
            mknod("config/NaviX.cnf")
            settings = Settings.get_instance(self.logger)
            return_value = settings.read_navix_config(self.logger, "DEBUG")
            # Test return value
            self.assertEqual(None, return_value)
            sleep(1)  # Time for logging
            with open(self.log_file_s, "r") as log_file:
                lines = log_file.readlines()  # Read to end
            log_file.close()
            with open('tests/test_configs/expected_log_empty_cnf_file.txt', "r") as log_file:
                ref_lines = log_file.readlines()  # Read to end
            log_file.close()
            lines = lines[-len(ref_lines):]
            # Test log output
            self.assertListEqual(sorted(ref_lines), sorted(lines))
        finally:
            rename("config/NaviX.cnf.bak", "config/NaviX.cnf")

    def test_empty_known_servers_config(self):
        # Check empty known_servers.yml
        Settings.get_instance(self.logger)._drop()
        rename("config/known_servers.yml", "config/known_servers.yml.bak")
        try:
            mknod("config/known_servers.yml")
            with self.assertRaises(AttributeError):
                settings = Settings.get_instance(self.logger)
                settings.read_navix_config(self.logger, "DEBUG")
            sleep(1)  # Time for logging
            with open(self.log_file_s, "r") as log_file:
                lines = log_file.readlines()  # Read to end
            log_file.close()
            with open('tests/test_configs/expected_log_empty_known_servers_file.txt', "r") as log_file:
                ref_lines = log_file.readlines()  # Read to end
            log_file.close()
            lines = lines[-len(ref_lines):]
            # Test log output
            self.assertListEqual(sorted(ref_lines), sorted(lines))
        finally:
            rename("config/known_servers.yml.bak", "config/known_servers.yml")

    def test_empty_htcondor_bindings_config(self):
        # Check empty HTCondor_bindings.cnf
        Settings.get_instance(self.logger)._drop()
        rename("config/HTCondor_bindings.cnf", "config/HTCondor_bindings.cnf.bak")
        try:
            mknod("config/HTCondor_bindings.cnf")
            settings = Settings.get_instance(self.logger)
            return_value = settings.read_htcondor_bindings_config(self.logger)
            # Test return value
            self.assertEqual(None, return_value)
            sleep(1)  # Time for logging
            with open(self.log_file_s, "r") as log_file:
                lines = log_file.readlines()  # Read to end
            log_file.close()
            with open('tests/test_configs/expected_log_empty_htcondor_bindings_file.txt', "r") as log_file:
                ref_lines = log_file.readlines()  # Read to end
            log_file.close()
            lines = lines[-len(ref_lines):]
            # Test log output
            self.assertListEqual(sorted(ref_lines), sorted(lines))
        finally:
            rename("config/HTCondor_bindings.cnf.bak", "config/HTCondor_bindings.cnf")

    # TODO: Write test for wrong input in all three configs

    def test_bad_argument(self):
        Settings.get_instance(self.logger)._drop()
        settings = Settings.get_instance(self.logger, logging_level="WRONG")
        self.assertEqual(settings.log_level, "DEBUG")




